Attribute VB_Name = "modSpectrumFFT"
Option Explicit
Private Const conPi = 3.14159265358979

Public Type TComplex
    re As Single
    im As Single
End Type

Public Sub FFTComplex(ByVal lngSamples As Long, ByVal blnInverseTransform As Boolean, _
ByRef udtComplexIn() As TComplex, ByRef udtComplexOut() As TComplex)
    Dim dblAngleNumerator As Double
    Dim dblDeltaAngle As Double, dblDeltaAr As Double, dblAlpha As Double, dblBeta As Double
    Dim ar As Double, ai As Double, tr, ti As Double
    Dim lngBlockSize As Long, lngBlockEnd As Long
    Dim lngNumBits As Long
    Dim i, j, k, n As Long
    
    If Not IsPowerOfTwo(lngSamples) Then Exit Sub
    ReDim udtComplexOut(UBound(udtComplexIn))
    
    If blnInverseTransform Then
        dblAngleNumerator = -(CDbl(2) * conPi)
    Else
        dblAngleNumerator = (CDbl(2) * conPi)
    End If
    
    lngNumBits = NumberOfBitsNeeded(lngSamples)
    
    For i = CLng(0) To lngSamples - CLng(1)
        j = ReverseBits(i, lngNumBits)
        udtComplexOut(j).re = udtComplexIn(i).re
        udtComplexOut(j).im = udtComplexIn(i).im
    Next i
    
    lngBlockEnd = CLng(1)
    lngBlockSize = CLng(2)
    
    Do While lngBlockSize <= lngSamples
        dblDeltaAngle = dblAngleNumerator / CDbl(lngBlockSize)
        dblAlpha = Sin(CDbl(0.5) * dblDeltaAngle)
        dblAlpha = CDbl(2) * (dblAlpha ^ CDbl(2))
        dblBeta = Sin(dblDeltaAngle)
        
        i = CLng(0)
        
        Do While i < lngSamples
            ar = CDbl(1)
            ai = CDbl(0)
            
            j = i
            
            For n = CLng(0) To lngBlockEnd - CLng(1)
                k = j + lngBlockEnd
                
                tr = ar * udtComplexOut(k).re - ai * udtComplexOut(k).im
                ti = ai * udtComplexOut(k).re + ar * udtComplexOut(k).im
                
                udtComplexOut(k).re = udtComplexOut(j).re - tr
                udtComplexOut(k).im = udtComplexOut(j).im - ti
                
                udtComplexOut(j).re = udtComplexOut(j).re + tr
                udtComplexOut(j).im = udtComplexOut(j).im + ti
                
                dblDeltaAr = dblAlpha * ar + dblBeta * ai
                
                ai = ai - (dblAlpha * ai - dblBeta * ar)
                ar = ar - dblDeltaAr
                
                j = j + CLng(1)
            Next n
            
            i = i + lngBlockSize
        Loop
        
        lngBlockEnd = lngBlockSize
        lngBlockSize = lngBlockSize * CLng(2)
    Loop
    
    If blnInverseTransform Then
        For i = CLng(0) To lngSamples - CLng(1)
            udtComplexOut(i).re = udtComplexOut(i).re / CDbl(lngSamples)
            udtComplexOut(i).im = udtComplexOut(i).im / CDbl(lngSamples)
        Next i
    End If
End Sub

Private Function IsPowerOfTwo(ByVal lngValue As Long) As Boolean
    If lngValue < CLng(2) Then
        IsPowerOfTwo = False
        Exit Function
    Else
        IsPowerOfTwo = (lngValue And (lngValue - CLng(1)) <> CLng(0))
    End If
End Function

Private Function NumberOfBitsNeeded(ByVal lngSamples As Long) As Long
    Dim lngByteIndex As Byte
    
    For lngByteIndex = CLng(0) To CLng(16)
        If (lngSamples And (CLng(2) ^ lngByteIndex)) <> CLng(0) Then
            NumberOfBitsNeeded = lngByteIndex
            Exit Function
        End If
    Next lngByteIndex
End Function

Private Function ReverseBits(ByVal lngIndex As Long, ByVal lngBits As Long) As Long
    Dim lngBitIndex As Long
    Dim lngReverse As Long
    
    lngReverse = CLng(0)
    
    For lngBitIndex = CLng(0) To lngBits - CLng(1)
        lngReverse = (lngReverse * CLng(2)) Or (lngIndex And CLng(1))
        lngIndex = lngIndex \ CLng(2)
    Next lngBitIndex
    
    ReverseBits = lngReverse
End Function

Private Function IndexToFrequency(ByVal lngBaseFrequency As Long, ByVal lngSamples As Long, _
ByVal lngIndex As Long) As Double
    If lngIndex >= lngSamples Then
        IndexToFrequency = CDbl(0)
    ElseIf CDbl(lngIndex) <= CDbl(lngSamples) / CDbl(2) Then
        IndexToFrequency = (CDbl(lngIndex) / CDbl(lngSamples)) * CDbl(lngBaseFrequency)
    Else
        IndexToFrequency = (-(CDbl(lngSamples) - CDbl(lngIndex)) / CDbl(lngSamples)) * CDbl(lngBaseFrequency)
    End If
End Function

Private Function DoubleArrayIsEmpty(ByRef dblArray() As Double) As Boolean
    On Local Error GoTo ErrHandler
    Dim lngUBound As Long
    
    lngUBound = UBound(dblArray())
    DoubleArrayIsEmpty = False
    On Local Error GoTo 0
    Exit Function
    
ErrHandler:
    Err.Clear
    On Local Error GoTo 0
    DoubleArrayIsEmpty = True
End Function
